#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>
#include <mpi.h>
#include <memory.h>
#include <unistd.h>

#include "log.H"
#include "list.H"
#include "tsp.H"

int myrank, NumProcs, NumCities;

int *Dist;

Path::Path()
{
  length=0;
  visited=1;
  for (int i=0;i < NumCities;i++) city[i]=i;
}

void Path::Set(int len, int *cit, int vis)
{
  length = len;
  memcpy(city, cit, NumCities*sizeof(int));
  visited = vis;
}

void Path::Print()
{
  for(int i=0; i<visited; i++) 
     printf("  %d", city[i]);
  printf("; length = %d\n", length);
}

void FillDist()
{
    debug("Fill dist called from rank %d", myrank);

    // If rank 0
    if (myrank == 0) {
	// Read number of cities from stdin
	if (scanf("%d", &NumCities) < 1)
	    fatal("Failed to read number of cities");
	assert(NumCities <= MAXCITIES);

	// Allocate matrix
	Dist = new int[NumCities * NumCities];

	// Read city distances from stdin
        if (myrank == 0)
	    for(int i = 0;i < NumCities;i++)
		for(int j = 0;j < NumCities;j++)
		    if (scanf("%d", &Dist[i * NumCities + j]) < 1)
			fatal("Failed to read distance");

	// Print number of cities and distances
	printf("Number of cities: %d\n", NumCities);
	for(int i = 0;i < NumCities; i++) {
	    for(int j=0;j < NumCities; j++)
		printf("%5d", Dist[i * NumCities + j]);
	    printf("\n");
	}
    } else {
	// Rank not 0
	MPI_Status status;

	// Get number of cities
	if (MPI_Send(NULL, 0, MPI_INT, 0, GET_NUM_CITIES_TAG, MPI_COMM_WORLD) != MPI_SUCCESS)
	    fatal("Worker %d has failed to query for number of cities", myrank);
	if (MPI_Recv(&NumCities, 1, MPI_INT, 0, GET_NUM_CITIES_TAG, MPI_COMM_WORLD, &status) == MPI_SUCCESS)
	    debug("Worker %d has obtained number of cities: %d", myrank, NumCities);
	    else fatal("Worker %d has failed to obtain number of cities", myrank);

	// Allocate distance matrix
	Dist = new int[NumCities * NumCities];

	// Request distance matrix
	if (MPI_Send(NULL, 0, MPI_INT, 0, GET_DIST_MATRIX_TAG, MPI_COMM_WORLD) != MPI_SUCCESS)
	    fatal("Worker %d has failed to query for dist matrix", myrank);
	if (MPI_Recv(Dist, NumCities * NumCities, MPI_INT, 0, GET_DIST_MATRIX_TAG, MPI_COMM_WORLD, &status) == MPI_SUCCESS)
	    debug("Worker %d has obtained dist matrix", myrank);
	else fatal("Worker %d has failed to obtain dist matrix", myrank);
    }
}

void Coordinator()
{
    MPI_Status status;
    Msg_t msg;

    int* waiting = new int[NumProcs];
    int nwait = 0;
    int bpath = 0;

    Path Shortest;
    LIST queue;
    Path *path = new Path;
    queue.Insert(path, 0);

    Shortest.length = INT_MAX;

    debug("Coordinator started");

    while (nwait < NumProcs-1) {
	// Blocking receive
	MPI_Recv(&msg, MSGSIZE, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
	debug("Coordinator got message: tag=%d, size=%d", status.MPI_TAG, MSGSIZE);

	switch (status.MPI_TAG)
	{
	    case GET_NUM_CITIES_TAG:
		MPI_Send(&NumCities, 1, MPI_INT, status.MPI_SOURCE, GET_NUM_CITIES_TAG, MPI_COMM_WORLD);
		break;
	    case GET_DIST_MATRIX_TAG:
		MPI_Send(Dist, NumCities * NumCities, MPI_INT, status.MPI_SOURCE, GET_DIST_MATRIX_TAG, MPI_COMM_WORLD);
		break;
	    case BEST_PATH_TAG:
		if (msg.length < Shortest.length) {
		    bpath++;
		    debug("Got best path %d, source = %d, length = %d", bpath, status.MPI_SOURCE, msg.length);

		    Shortest.Set(msg.length, msg.city, NumCities);
		    for(int i = 1;i < NumProcs;i++)
		    MPI_Send(&(Shortest.length), 1, MPI_INT, i, UPDATE_BEST_PATH_TAG, MPI_COMM_WORLD);
		}
		break;
	    case PUT_PATH_TAG:
		if (nwait > 0)
		    MPI_Send(&msg, MSGSIZE, MPI_INT, waiting[--nwait], REPLY_PATH_TAG, MPI_COMM_WORLD);
		else {
		    path=new Path();
		    path->Set(msg.length, msg.city, msg.visited);
		    queue.Insert(path, msg.length);
		}
		break;
	    case GET_PATH_TAG:
		if (! queue.IsEmpty()) {
		    path=(Path *) queue.Remove(NULL);
		    msg.length = path->length;
		    memcpy(msg.city, path->city, MAXCITIES * sizeof(int));
		    msg.visited = path->visited;
		    MPI_Send(&msg, MSGSIZE, MPI_INT, status.MPI_SOURCE, REPLY_PATH_TAG, MPI_COMM_WORLD);
		    delete path;
		} else {
		    waiting[nwait++] = status.MPI_SOURCE;
		    if (nwait == NumProcs-1)
		    for (int i=1;i < NumProcs;i++)
			MPI_Send(NULL, 0, MPI_INT, i, DONE_TAG, MPI_COMM_WORLD);
		}
		break;
	}
    }
    printf("Shortest path:\n");
    Shortest.Print();
}

void Worker() {
    MPI_Status status;
    Msg_t msg;
    int shortestLength = INT_MAX;

    debug("Worker %d started", myrank);

    // Request path tag
    MPI_Send(NULL, 0, MPI_INT, 0, GET_PATH_TAG, MPI_COMM_WORLD);

    while(1) {
	// Receive
	MPI_Recv(&msg, MSGSIZE, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

	// Process done tag
	if (status.MPI_TAG == DONE_TAG) {
	    debug("Worker %d received DONE_TAG", myrank);
	    break;
	}

	// Update best path tag
	if (status.MPI_TAG == UPDATE_BEST_PATH_TAG) {
	    debug("Worker %d received UPDATE_BEST_PATH_TAG to %d", myrank, msg.length);
	    shortestLength = msg.length;
	    continue;
	}

	// Increase number of visited cities
	msg.visited++;
	if (msg.visited == NumCities) {
	    int d1 = Dist[(msg.city[NumCities - 2]) * NumCities + msg.city[NumCities - 1]];
	    int d2 = Dist[(msg.city[NumCities - 1]) * NumCities];

	    if (d1 * d2) {
		msg.length += d1 + d2;
		if (msg.length < shortestLength)
		    MPI_Send(&msg, MSGSIZE, MPI_INT, 0, BEST_PATH_TAG, MPI_COMM_WORLD);
	    }
	} else {
	    int length = msg.length;
	    for(int i=msg.visited - 1;i < NumCities;i++) {
		if (i > msg.visited - 1) {
		    int tmp = msg.city[msg.visited - 1];
		    msg.city[msg.visited - 1] = msg.city[i];
		    msg.city[i] = tmp;
		}

		if (int d = Dist[(msg.city[msg.visited - 2]) * NumCities + msg.city[msg.visited-1]]) {
		    msg.length = length + d;
		    if (msg.length < shortestLength)
			MPI_Send(&msg, MSGSIZE, MPI_INT, 0, PUT_PATH_TAG, MPI_COMM_WORLD);
		}
	    }
	}

	MPI_Send(NULL, 0, MPI_INT, 0, GET_PATH_TAG, MPI_COMM_WORLD);
    }
    debug("Worker %d finished", myrank);
}

void ProcessCommandLine(int argc, char *argv[]) {
    int c;
    while((c=getopt(argc, argv, "d")) != -1)
	switch(c) {
	    case 'd':
		SetDebugReported(true);
		break;
	    default:
		exit(1);
		break;
	}
}

int main(int argc, char *argv[])
{
    ProcessCommandLine(argc, argv);

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
    MPI_Comm_size(MPI_COMM_WORLD, &NumProcs);

    if (NumProcs < 2)
	fatal("MPI reports %d procs available. Probably MPI is not bootstraped properly.", NumProcs);

    debug("Started worker %d", myrank);

    FillDist();

    if (myrank == 0)
	Coordinator();
    else Worker();

    MPI_Finalize();

  return 0;
}
